/****************************************************************************
 * @file DS1302.c - MAXIM DS1302 Real Time Clock library implementation file
 * Copyright (c) 2009, Craig Smith
 * See end of file for MIT license terms.
 *
 * This is a library of routines to interface a MAXIM DS1302 real time
 * clock with a Parallax Propeller I microcontroller.
 *
 * The code was developed using ImageCraft's Propeller C (ICCv7) ver. 7.04
 ****************************************************************************/

#include <propeller.h>
#include "DS1302.h"

/* Prototypes for static functions at end of this file */
static void write(BYTE cmd, BYTE data);
static void writeByte(BYTE cmd);
static BYTE read(BYTE cmd);
static BYTE readByte(void);
static BYTE convert1302ToDec(BYTE dataIn);
static BYTE convertDecTo1302(BYTE dataIn);
void smallDelay(void);

/* static variables holding the pins used to communicate with the DS1302 */
static unsigned long pin_clk, pin_io, pin_ce;
  
/****************************************************************************
 * Initialize variables and pins and configure default DS1302 options
 *   inClk = Prop pin tied to DS1302 clock pin  
 *   inIo  = Prop pin tied to DS1302 io pin
 *   inCE  = Prop pin tied to DS1302 ce pin
 ****************************************************************************/
void ds1302_init(BYTE clkPin, BYTE ioPin, BYTE cePin)
{
  pin_clk  = 1 << clkPin;             // initialize pin variables
  pin_io  = 1 << ioPin;
  pin_ce  = 1 << cePin;

  DIRA |=  pin_ce;                    // Set DS1302 CE pin to output
  OUTA &= ~pin_ce; 					  // Set DS1302 CE pin low
  smallDelay();                       
  DIRA |=  pin_clk;                   // Set DS1302 CLK pin to output
  OUTA &= ~pin_clk; 				  // Set DS1302 CLK pin low
  smallDelay();
  write(0x90,0x00);                   // Disable charging register 
  write(0x8E,0x00);                   // Clear write-protect bit
  write(0x84,0x00);                   // Set to 24 hour mode
  write(0x80,0x00);                   // clear clock halt
}

/****************************************************************************
 * Set date and time in DS1302
 *  dt is ds1302_structure containing date and time information
 *  see ds1320.h for structure declaration
 ****************************************************************************/
void ds1302_setDatetime(struct ds1302_dateTime dt)
{
  BYTE hour, tc;
  
  hour = convertDecTo1302(dt.hour);
  if (dt.is12HrMode)
    {
	if (dt.hour>12)
      hour = convertDecTo1302(dt.hour-12);
	SET(hour, _1224BIT);                // set 12/24 hour mode bit
    if (dt.pm)  
      SET(hour, AMPMBIT);               // set AM/PM bit
	}
  else
	CLR(hour, _1224BIT);              // clear 12/24 hour mode bit
  tc = ds1302_getTrickleCharge();

  OUTA |= pin_ce;                          // toggle chip enable high      
  writeByte( 0xBE );                       // send command byte
  writeByte(convertDecTo1302(dt.second));  // must write 8 bytes for DS1302
  writeByte(convertDecTo1302(dt.minute));  // to commit data
  writeByte(hour);  
  writeByte(convertDecTo1302(dt.day));  
  writeByte(convertDecTo1302(dt.month));  
  writeByte(dt.dow);  
  writeByte(convertDecTo1302(dt.year));
  writeByte(tc);                           // retain trickle charge setting
  OUTA &= ~pin_ce;                         // toggle chip enable low
}

/****************************************************************************
 * Get date and time from DS1302
 *  dt is pointer to ds1302_structure to receive date and time information
 *  see ds1320.h for structure declaration
 ****************************************************************************/
void ds1302_getDatetime(struct ds1302_dateTime *dt)
{
  
  BYTE hour;
  
  OUTA |= pin_ce;               // toggle chip enable high      
  writeByte( 0xBF );            // send command byte for clock burst mode read
  dt->second = convert1302ToDec(readByte());  
  dt->minute = convert1302ToDec(readByte());  
  hour = readByte();  
  dt->day = convert1302ToDec(readByte());  
  dt->month = convert1302ToDec(readByte());  
  dt->dow = readByte();  
  dt->year = convert1302ToDec(readByte());
  OUTA &= ~pin_ce;              // toggle chip enable low

  if (CHECKBIT(hour, _1224BIT))
    {                           // DS1302 is in 12 hour mode
	dt->is12HrMode = TRUE;
	dt->pm = CHECKBIT(hour, AMPMBIT);     
	CLEARBITMASK(hour, 0xA0);   // strip upper bits to get hour data alone
	}
  else
    {
	dt->is12HrMode = FALSE;     // DS1302 is not in 12 hour mode
	dt->pm = FALSE;
	CLR(hour, _1224BIT);   // strip upper bits to get hour data alone
	}
  dt->hour = convert1302ToDec(hour);
}

/****************************************************************************
 * Set calendar day in DS1302
 *  day = 1 to 31
 ****************************************************************************/
void ds1302_setDay(BYTE day)
{
  write(0x86, convertDecTo1302(day));;
}
  
/****************************************************************************
 * Get calendar day from DS1302
 *  returns day as 1 to 31
 ****************************************************************************/
 BYTE ds1302_getDay(void)
{
  return(convert1302ToDec(read(0x87)));
}

/****************************************************************************
 * Set calendar month in DS1302
 *  month = 1 to 12
 ****************************************************************************/
void ds1302_setMonth(BYTE month)
{
  write(0x88, convertDecTo1302(month));;
}

/****************************************************************************
 * Get calendar month from DS1302
 *  returns month as 1 to 12
 ****************************************************************************/
BYTE ds1302_getMonth(void)
{
  return(convert1302ToDec(read(0x89)));
}

/****************************************************************************
 * Set calendar year in DS1302
 *  month = 0 to 99
 ****************************************************************************/
void ds1302_setYear(BYTE year)
{
  write(0x8c, convertDecTo1302(year));;
}

/****************************************************************************
 * Get calendar year from DS1302
 *  returns year as 0 to 99
 ****************************************************************************/
BYTE ds1302_getYear(void)
{
  return(convert1302ToDec(read(0x8d)));
}
 
/****************************************************************************
 * Set day of week in DS1302
 *  dow = 1 to 7
 *  You define 1 to mean any day you want.  Subsequent days increase
 *  by one.
 ****************************************************************************/
void ds1302_setDOW(BYTE dow)
{
  write(0x8a, convertDecTo1302(dow));;
}

/****************************************************************************
 * Get dow of week from DS1302
 *  returns day as 1 to 7
 ****************************************************************************/
BYTE ds1302_getDOW(void)
{
  return(convert1302ToDec(read(0x8B)));
}

/****************************************************************************
 * Set hour in DS1302
 *  hour = 0 to 23 in 24 hour mode (default)
 *  hour = 1 to 12 in 12 hour mode
 *  AM or PM can be set using ds1302_setPM()
 ****************************************************************************/
void ds1302_setHour(BYTE hour)
{
  hour = convertDecTo1302( hour );
  if (ds1302_get12HourMode())
    {
	SET(hour, _1224BIT);                // set 12/24 mode bit
    if (ds1302_getPM())
	  SET(hour, AMPMBIT);
	}	 					               // retain AM/PM bit
  write(0x84, hour);
}

/****************************************************************************
 * Get hour from DS1302
 *  returns hour as 0 to 23 in 24 hour mode (default)
 *  returns hour as 1 to 12 in 12 hour mode
 *  AM or PM can be retrieved using ds1302_getPM()
 ****************************************************************************/
BYTE ds1302_getHour(void)
{
  BYTE hr;
  
  hr = read(0x85) & ~0x80;  			  // strip 12/24 mode bit
  if (ds1302_get12HourMode())
    CLR(hr, AMPMBIT);                 // strip AM/PM bit
  return(convert1302ToDec(hr));
}

/****************************************************************************
 * Set minute in DS1302
 *  minute = 0 to 59
 ****************************************************************************/
void ds1302_setMinute(BYTE minute)
{
  write(0x82, convertDecTo1302(minute));;
}

/****************************************************************************
 * Get minute from DS1302
 *  returns minute as 0 to 59
 ****************************************************************************/
BYTE ds1302_getMinute(void)
{
  return(convert1302ToDec(read(0x83)));
}

/****************************************************************************
 * Set second in DS1302
 *  second = 0 to 59
 ****************************************************************************/
void ds1302_setSecond(BYTE second)
{
  write(0x80, convertDecTo1302(second));;
}

/****************************************************************************
 * Get second from DS1302
 *  returns second as 0 to 59
 ****************************************************************************/
BYTE ds1302_getSecond(void)
{
  return(convert1302ToDec(read(0x81)));
} 

/****************************************************************************
 * Set AM or PM clock mode of DS1302
 *   pm = TRUE if PM, FALSE if AM
 *
 * NOTE:  This function has no effect if clock is not in 12 hour mode 
 ****************************************************************************/
void ds1302_setPM(BYTE pm)
{
  if (ds1302_get12HourMode())
  {
    if (pm)
  	   write(0x84, ((read(0x85)) | 0x20));  // retain hour and 12/24 mode bit
    else
  	   write(0x84, ((read(0x85)) & ~0x20));
  }
}

/****************************************************************************
 * Get AM or PM clock mode of DS1302
 *  returns TRUE if PM, FALSE if AM 
 ****************************************************************************/
BYTE ds1302_getPM(void)
{
  if (ds1302_get12HourMode())
    return(read(0x85) & 0x20);
  else
    return 0;
}

/****************************************************************************
 * Set 12 or 24 hour clock mode of DS1302
 *   goto12HourMode = TRUE if 12 hour, FALSE if 24 hour (default) 
 ****************************************************************************/
void ds1302_set12HourMode(BYTE goto12HourMode)
{
  BYTE currentHour;
  
  if (ds1302_get12HourMode())
  {                                         // it's in 12 hour mode
    if (goto12HourMode)   
      return;                              	// already in 12hr mode, do nothing
	else
	{                                       // turn off 12 hour mode
	  currentHour = ds1302_getHour();
	  if (ds1302_getPM())
	    currentHour += 12;                  // adjust hour for PM
	  write(0x84, convertDecTo1302(currentHour)); // write hour - other bits 0
	}
  }
  else 
  {                                         // it's in 24 hour mode
	if (goto12HourMode)                     
	{  
	  currentHour = ds1302_getHour(); 
      if (currentHour > 12)                 
	  {                                     // switching to 12 hour mode 
	    currentHour -= 12;                  // subtract 12 from hour if required
	    currentHour = convertDecTo1302(currentHour);
		SET(currentHour, _1224BIT);
		SET(currentHour, AMPMBIT);		
	    write(0x84, currentHour); // update hour, set 12hr and AM/PM
	  }
	  else
	  {
	    SET(currentHour, _1224BIT);
        write(0x84, currentHour); // retain hour, set 12hr mode bit
	  }
    }
	else
	  return;                              // already in 24hr mode, do nothing
  }
}

/****************************************************************************
 * Get 12 or 24 hour clock mode of DS1302
 *   returns TRUE if 12 hour, FALSE if 24 hour 
 ****************************************************************************/
BYTE ds1302_get12HourMode(void)
{
  return(CHECKBIT(read(0x85), _1224BIT));               // set bit 7
}

/****************************************************************************
 * Set write protect option of DS1302
 *   on = TRUE if write protect, FALSE if not (default)  
 ****************************************************************************/
void ds1302_setWriteProtect(BYTE on)
{
  if (on)
    write(0x8E,0x80);                      // set bit 7 
  else
    write(0x8E,0x00);                      // clear bit 7      
}

/****************************************************************************
 * Get write protect option of DS1302
 *   return TRUE if write protect, FALSE if not (default)  
 ****************************************************************************/
BYTE ds1302_getWriteProtect(void)
{
  return (read(0x8F)); 
}

/****************************************************************************
 * Set clock halt option of DS1302
 *   halt = TRUE if clock halted, FALSE if not (default)   
 ****************************************************************************/
void ds1302_setClockHalt(BYTE halt)
{
  if (halt)
    write(0x80, (read(0x81) | 0x80));      // set bit 7 
  else
    write(0x80, (read(0x81) & 0x7F));      // clear bit 7
}

/****************************************************************************
 * Get clock halt option of DS1302
 *   return TRUE if clock halted, FALSE if not (default)   
 ****************************************************************************/
BYTE ds1302_getClockHalt(void)
{
  return (read(0x81)); 
}

/****************************************************************************
 * Set trickle charge option of DS1302
 *   mode = 0 to turn off trickle charge (default)  
 *          1 for 1 diode and 2k ohm resistor
 *          2 for 1 diode and 4k ohm resistor
 *          3 for 1 diode and 8k ohm resistor
 *          4 for 2 diodes and 2k ohm resistor
 *          5 for 2 diodes and 4k ohm resistor
 *          6 for 2 diodes and 8k ohm resistor
 ****************************************************************************/
void ds1302_setTrickleCharge(BYTE mode)
{
  switch (mode)
  {
  case 0:
       write(0x90,0x00);
	   break; 
  case 1:
       write(0x90,0xA5);
	   break; 
  case 2:
       write(0x90,0xA6);
	   break; 
  case 3:
       write(0x90,0xA7);
	   break; 
  case 4:
       write(0x90,0xA9);
	   break; 
  case 5:
       write(0x90,0xAA);
	   break; 
  case 6:
       write(0x90,0xAB);
	   break; 
  }
}

/****************************************************************************
 * Get trickle charge option of DS1302
 *   return 0 for disabled trickle charge
 *          1 for 1 diode and 2k ohm resistor
 *          2 for 1 diode and 4k ohm resistor
 *          3 for 1 diode and 8k ohm resistor
 *          4 for 2 diodes and 2k ohm resistor
 *          5 for 2 diodes and 4k ohm resistor
 *          6 for 2 diodes and 8k ohm resistor
 ****************************************************************************/
BYTE ds1302_getTrickleCharge(void)
{
  BYTE mode;
  
  mode = read(0x91);
  switch (mode)
  {
  case 0:
       return(0);
	   break; 
  case 0xA5:
       return(1);
	   break; 
  case 0xA6:
       return(2);
	   break; 
  case 0xA7:
       return(3);
	   break; 
  case 0xA9:
       return(4);
	   break; 
  case 0xAA:
       return(5);
	   break; 
  case 0xAB:
       return(6);
	   break; 
  default:
       return(0);
  }
}

/****************************************************************************
 * Set byte in DS1302 RAM
 *   address = 0 to 30
 *   data = 0 to 255
 ****************************************************************************/
void ds1302_setRamByte(BYTE address, BYTE data)
{
  if (address<30)
    write(0xC0 | (address<<1), data);
}
 
/****************************************************************************
 * Get byte from DS1302 RAM
 *  address = 0 to 30	
 *  returns data byte 
 ****************************************************************************/
BYTE ds1302_getRamByte(BYTE address)
{
  if (address<30)
    return(read(0xC1 | (address<<1)));
  else
    return(0);
}

/****************************************************************************
 * Set a block of RAM in DS1302
 *  bytes is an array of up to 31 bytes
 *  bytes is an array of up to 31 bytes
  ****************************************************************************/
void ds1302_setRamBlock(BYTE address, BYTE length, BYTE *data)
{
  BYTE i;
  
  OUTA |= pin_ce;               // toggle chip enable high      
  writeByte( 0xFE );            // send command byte
  for (i=address; i<(address+length); i++)
    writeByte(*data++);  
  OUTA &= ~pin_ce;              // toggle chip enable low
}

/****************************************************************************
 * Get a block of RAM from DS1302
 *  dt is pointer to ds1302_structure to receive date and time information
 *  see ds1320.h for structure declaration
 ****************************************************************************/
void ds1302_getRamBlock(BYTE address, BYTE length, BYTE *data)
{
  BYTE i;
  
  OUTA |= pin_ce;               // toggle chip enable high      
  writeByte( 0xFF );            // send command byte for clock burst mode read
  for (i=address; i<(address+length); i++)
    *data++ = readByte();  
  OUTA &= ~pin_ce;              // toggle chip enable low
}

/* static low-level functions follow ****************************************/

/* write a byte by clocking it out to the DS1302 */
static void writeByte(BYTE data)
{
  BYTE i, bit;
   
  DIRA |=  pin_io;        // set DS1302 IO pin to output
  bit = 0x01;             // initialize bit mask
  for (i=0; i<=7; i++)    // clock out instructions
  {
    if ( data & bit ) 	  // test data bit
	  OUTA |= pin_io;     // set DS1302 IO pin high
    else 
	  OUTA &= ~pin_io;    // set DS1302 IO pin low
    smallDelay();         // pause at least 2 us for 'CLK to Data' delay
    OUTA |= pin_clk;      // toggle clock pin high        
    bit <<= 1;            // shift mask for next bit ('CLK to Data' delay also)
    OUTA &= ~pin_clk;     // toggle clock pin low
	smallDelay();         // pause at least 2 us for 'CLK to Data' delay
  }
}

static void write(BYTE cmd, BYTE data)
{
  OUTA |= pin_ce;               // toggle chip enable high        
  writeByte(cmd);               // send command byte
  writeByte(data);              // send data byte
  OUTA &= ~pin_ce;              // toggle chip enable low
}

static BYTE readByte(void)
{
  BYTE i, delay, datar;

  DIRA &= ~pin_io;              // toggle io pin low
  datar = 0;
  for (i=0; i<=7; i++)
  {
    if (INA & pin_io)
      datar += (1 << i);
    OUTA |= pin_clk;          // toggle clock high        
    smallDelay();             // pause at least 2 us for 'CLK to Data' delay
    OUTA &= ~pin_clk;         // toggle clock low
    smallDelay();             // pause at least 2 us for 'CLK to Data' delay
  }
  return(datar);
}

/* read a byte by clocking it in from the DS1302 */
static BYTE read(BYTE cmd)
{
  BYTE i, delay, datar;

  OUTA |= pin_ce;               // toggle chip enable high        
  writeByte(cmd);               // send command byte
  DIRA &= ~pin_io;              // toggle io pin low
  datar = 0;
  for (i=0; i<=7; i++)
  {
    if (INA & pin_io)
      datar += (1 << i);
    OUTA |= pin_clk;          // toggle clock high        
    smallDelay();             // pause at least 2 us for 'CLK to Data' delay
    OUTA &= ~pin_clk;         // toggle clock low
    smallDelay();             // pause at least 2 us for 'CLK to Data' delay
  }
  OUTA &= ~pin_ce;              // toggle chip enable low
  return(datar);
}

static BYTE convert1302ToDec(BYTE data)
{
  return((data & 0x0f) + ((data/16) & 0x0f) * 10);
}

static BYTE convertDecTo1302(BYTE data)
{
  BYTE tmp;
  
  tmp = data/10;
  return(data - (tmp * 10) + (tmp * 16));
}

void smallDelay(void)
{
  asm("nop");
  asm("nop");
}

/*
+-----------------------------------------------------------------------------+
                         TERMS OF USE: MIT License                           
+-----------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy 
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights 
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell    
copies of the Software, and to permit persons to whom the Software is        
furnished to do so, subject to the following conditions:                     
                                                                             
 The above copyright notice and this permission notice shall be included in  
 all copies or substantial portions of the Software.                         
                                                                             
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER      
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     
 FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         
 DEALINGS IN THE SOFTWARE.                                                   
+-----------------------------------------------------------------------------+
*/